#include "defs.h"
#ifdef CILKPAR
#include <cilk.h>
#include <reducer_opadd.h>
#else
#define cilk_for for
#define cilk_main main
#define cilk_spawn
#define cilk_sync
#endif

#define MAX_THREADS 320

double betweennessCentrality(graph* G, double* BC) {
  int* Srcs; 
  int seed = 2387;
  double elapsed_time;
  int numV, n, m;
 	int mv;

  n = G->nv;
  m = G->ne;
  numV = n;

	// We use a reducer to accumulate the betweeness measure between runs of BFS
	typedef cilk::reducer_opadd<double> bc_reducer;
	bc_reducer *bc_reducers = new bc_reducer[ numV ];

  /* Permute vertices */
  Srcs = (int *) malloc(n*sizeof(int));
  for ( int i=0; i<n; i++) {
    Srcs[i] = i;
		bc_reducers[i].set_value(0);
  }

  /* Start timing code from here */
  elapsed_time = get_seconds();

  int *in_degree = (int *) calloc(n+1, sizeof(int));
	for ( int i=0; i<m; i++) {
	  mv = G->nbr[i];
    in_degree[mv]++;
	}
  
  /***********************************/
  /*** MAIN LOOP *********************/
  /***********************************/
	cilk_for ( int p = 0; p<MAX_THREADS; p++ ) {
		
		/* local memory for each thread */  
	  int num_traversals = 0;
	  int myCount = 0;
		int i, count, phase_num;
		
	  int *S   = (int *) malloc(n*sizeof(int));
	  double *sig = (double *) malloc(n*sizeof(double));
	  int *d   = (int *) malloc(n*sizeof(int));
	  double *del = (double *) calloc(n, sizeof(double));
	
	  int *start = (int *) malloc(n*sizeof(int));
	  int *end = (int *) malloc(n*sizeof(int));
		
		/* Initialize predecessor lists */
	  /* Number of predecessors of a vertex is at most its in-degree. */
	  plist* P = (plist *) calloc(n, sizeof(plist));
	  int *numEdges = (int *) malloc((n+1)*sizeof(int));
	  
	  prefix_sums(in_degree, numEdges, n);
	
	  int *pListMem = (int *) malloc(m*sizeof(int));
	  for ( int i=0; i<n; i++) {
	    P[i].list = pListMem + numEdges[i];
	    P[i].degree = in_degree[i];
	    P[i].count = 0;
	  }
	  free(numEdges);

		// Graph Computation
	  for ( i=0; i<n; i++) {
	    d[i] = -1;
	  }
		
		for ( int r=p*n/MAX_THREADS; r<(p+1)*n/MAX_THREADS; r++ ) {

			i = Srcs[r]; // Currently counts up from 0, 1, etc
			if (G->firstnbr[i+1] - G->firstnbr[i] == 0) { // Its not connected
				continue;
			} else {
				num_traversals++;
			}

			if (num_traversals == numV + 1) { // How does this make any sense? n_ts never gets that high
				break;
			}
		
			sig[i] = 1;
			d[i] = 0;
			S[0] = i;
			start[0] = 0;
			end[0] = 1;
		
			int count = 1;
			int phase_num = 0;
			int v, w;

			while (end[phase_num] - start[phase_num] > 0) {
				myCount = 0;
				// BFS to destination, calculate distances, 
				for ( int vert = start[phase_num]; vert < end[phase_num]; vert++ ) {
					v = S[vert];
				
					for ( int j=G->firstnbr[v]; j<G->firstnbr[v+1]; j++ ) {
						w = G->nbr[j];
						if (v != w) {

							/* w found for the first time? */ 
							if (d[w] == -1) {
								//printf("n=%d, j=%d, start=%d, end=%d, count=%d, vert=%d, w=%d, v=%d\n",n,j,start[phase_num],end[phase_num],myCount,vert,w,v);
								S[end[phase_num] + myCount] = w;
								myCount++;
								d[w] = d[v] + 1; 
								sig[w] = sig[v]; 
								P[w].list[P[w].count++] = v;
							} else if (d[w] == d[v] + 1) {
								sig[w] += sig[v]; 
								P[w].list[P[w].count++] = v;
							}
						
						}
					}
	 			}
			
				/* Merge all local stacks for next iteration */
				phase_num++; 
				
				start[phase_num] = end[phase_num-1];
				end[phase_num] = start[phase_num] + myCount;
			
				count = end[phase_num];
			}
 	
			phase_num--;
					
			//Calculate the betweenness centrality
			while (phase_num > 0) {
				for ( int j=start[phase_num]; j<end[phase_num]; j++ ) {
					w = S[j];
					for ( int k = 0; k < P[w].count; k++ ) {
						v = P[w].list[k];
						del[v] = del[v] + sig[v]*(1+del[w])/sig[w];
					}
					bc_reducers[w] += del[w];
				}

				phase_num--;
			}
		
			// Reset everything for the next go
			for ( int j=0; j<count; j++ ) {
				w = S[j];
				d[w] = -1;
				del[w] = 0;
				P[w].count = 0;
			}
		}
		
		// Deallocate temporaries
  	free(S);
		free(sig);
	  free(d);
	  free(del);
	  free(start);
	  free(end);
		free(pListMem);
	  free(P);
  }
  /***********************************/
  /*** END OF MAIN LOOP **************/
  /***********************************/
 
	for( int i = 0; i<numV; i++ )
		BC[i] = bc_reducers[i].get_value();
  
  elapsed_time = get_seconds() - elapsed_time;

  free(Srcs);
  free(in_degree);

  return elapsed_time;
}


/*
 * Serial Version
 *
 */
double betweennessCentrality_serial(graph* G, double* BC) {
  int *S; 	/* stack of vertices in order of distance from s. Also, implicitly, the BFS queue */
  plist* P;  	/* predecessors of vertex v on shortest paths from s */
  double* sig; 	/* No. of shortest paths */
  int* d; 	/* Length of the shortest path between every pair */
  double* del; 	/* dependency of vertices */
  int *in_degree, *numEdges;
  int *pListMem;	
  int* Srcs; 
  int *start, *end;
  int seed = 2387;
  double elapsed_time;
  int i, j, k, p, count, myCount;
  int v, w, vert;
  int numV, num_traversals, n, m, phase_num;

  /* numV: no. of vertices to run BFS from = 2^K4approx */
  //numV = 1<<K4approx;
  n = G->nv;
  m = G->ne;
  numV = n;

  /* Permute vertices */
  Srcs = (int *) malloc(n*sizeof(int));
  for (i=0; i<n; i++) {
    Srcs[i] = i;
  }

  /* Start timing code from here */
  elapsed_time = get_seconds();

  /* Initialize predecessor lists */
  /* Number of predecessors of a vertex is at most its in-degree. */
  P = (plist *) calloc(n, sizeof(plist));
  in_degree = (int *) calloc(n+1, sizeof(int));
  numEdges = (int *) malloc((n+1)*sizeof(int));
  for (i=0; i<m; i++) {
    v = G->nbr[i];
    in_degree[v]++;
  }
  prefix_sums(in_degree, numEdges, n);
  pListMem = (int *) malloc(m*sizeof(int));
  for (i=0; i<n; i++) {
    P[i].list = pListMem + numEdges[i];
    P[i].degree = in_degree[i];
    P[i].count = 0;
  }
  free(in_degree);
  free(numEdges);
	
  /* Allocate shared memory */ 
  S   = (int *) malloc(n*sizeof(int));
  sig = (double *) malloc(n*sizeof(double));
  d   = (int *) malloc(n*sizeof(int));
  del = (double *) calloc(n, sizeof(double));
	
  start = (int *) malloc(n*sizeof(int));
  end = (int *) malloc(n*sizeof(int));

  num_traversals = 0;
  myCount = 0;

  for (i=0; i<n; i++) {
    d[i] = -1;
  }
	
  /***********************************/
  /*** MAIN LOOP *********************/
  /***********************************/
  for (p=0; p<n; p++) {

		i = Srcs[p];
		if (G->firstnbr[i+1] - G->firstnbr[i] == 0) {
			continue;
		} else {
			num_traversals++;
		}

		if (num_traversals == numV + 1) {
			break;
		}
		
		sig[i] = 1;
		d[i] = 0;
		S[0] = i;
		start[0] = 0;
		end[0] = 1;
		
		count = 1;
		phase_num = 0;

		while (end[phase_num] - start[phase_num] > 0) {
				myCount = 0;
				// BFS to destination, calculate distances, 
				for ( int vert = start[phase_num]; vert < end[phase_num]; vert++ ) {
					v = S[vert];
				
					for ( int j=G->firstnbr[v]; j<G->firstnbr[v+1]; j++ ) {
						w = G->nbr[j];
						if (v != w) {

							/* w found for the first time? */ 
							if (d[w] == -1) {
								//printf("n=%d, j=%d, start=%d, end=%d, count=%d, vert=%d, w=%d, v=%d\n",n,j,start[phase_num],end[phase_num],myCount,vert,w,v);
								S[end[phase_num] + myCount] = w;
								myCount++;
								d[w] = d[v] + 1; 
								sig[w] = sig[v]; 
								P[w].list[P[w].count++] = v;
							} else if (d[w] == d[v] + 1) {
								sig[w] += sig[v]; 
								P[w].list[P[w].count++] = v;
							}
						
						}
					}
	 			}
			
				/* Merge all local stacks for next iteration */
				phase_num++; 
				
				start[phase_num] = end[phase_num-1];
				end[phase_num] = start[phase_num] + myCount;
			
				count = end[phase_num];
		}
 	
		phase_num--;

		while (phase_num > 0) {
			for (j=start[phase_num]; j<end[phase_num]; j++) {
				w = S[j];
				for (k = 0; k < P[w].count; k++) {
					v = P[w].list[k];
					del[v] = del[v] + sig[v]*(1+del[w])/sig[w];
				}
				BC[w] += del[w];
			}

			phase_num--;
		}
		
		for (j=0; j<count; j++) {
			w = S[j];
			d[w] = -1;
			del[w] = 0;
			P[w].count = 0;
		}
  }
  /***********************************/
  /*** END OF MAIN LOOP **************/
  /***********************************/
 

	
  free(S);
  free(pListMem);
  free(P);
  free(sig);
  free(d);
  free(del);
  free(start);
  free(end);
  elapsed_time = get_seconds() - elapsed_time;
  free(Srcs);

  return elapsed_time;
}
